import pytest
from memcnn.data.sampling import NSamplesRandomSampler
import torch.utils.data as data
import numpy as np


@pytest.mark.parametrize('nsamples,data_samples', [(1, 1), (14, 10), (10, 14), (5, 1), (1, 5), (0, 10)])
def test_random_sampler(nsamples, data_samples):

    class TestDataset(data.Dataset):
        def __init__(self, elements):
            self.elements = elements

        def __getitem__(self, idx):
            return idx, idx

        def __len__(self):
            return self.elements

    datasrc = TestDataset(data_samples)
    sampler = NSamplesRandomSampler(datasrc, nsamples=nsamples)
    count = 0
    elements = []
    for e in sampler:
        elements.append(e)
        count += 1
        if count % data_samples == 0:
            assert len(np.unique(elements)) == len(elements)
            elements = []
    assert count == nsamples
    assert len(sampler) == nsamples
    assert sampler.__len__() == nsamples



import logging
import sys


def setup(use_stdout=True, filename=None, log_level=logging.DEBUG):
    """setup some basic logging"""

    log = logging.getLogger('')
    log.setLevel(log_level)
    fmt = logging.Formatter("%(asctime)s [%(name)-15s] %(message)s", datefmt="%y-%m-%d %H:%M:%S")

    if use_stdout:
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(log_level)
        ch.setFormatter(fmt)
        log.addHandler(ch)

    if filename is not None:
        fh = logging.FileHandler(filename)
        fh.setLevel(log_level)
        fh.setFormatter(fmt)
        log.addHandler(fh)



import pytest
import gc
import numpy as np
import math
from collections import defaultdict
import torch
import torch.nn
import memcnn.models.revop as revop
from memcnn.models.affine import AffineAdapterNaive


def readable_size(num_bytes):
    return '{:.2f}'.format(float(num_bytes) / float(1024 ** 2))


LEN = 79

# some pytorch low-level memory management constant
# the minimal allocate memory size (Byte)
PYTORCH_MIN_ALLOCATE = 2 ** 9
# the minimal cache memory size (Byte)
PYTORCH_MIN_CACHE = 2 ** 20


class MemReporter(object):
    """A memory reporter that collects tensors and memory usages
    Parameters:
        - model: an extra nn.Module can be passed to infer the name
        of Tensors
    """
    def __init__(self, model=None):
        self.tensor_name = defaultdict(list)
        self.device_mapping = defaultdict(list)
        self.device_tensor_stat = {}
        # to numbering the unknown tensors
        self.name_idx = 0
        if model is not None:
            assert isinstance(model, torch.nn.Module)
            # for model with tying weight, multiple parameters may share
            # the same underlying tensor
            for name, param in model.named_parameters():
                self.tensor_name[param].append(name)
        for param, name in self.tensor_name.items():
            self.tensor_name[param] = '+'.join(name)

    def _get_tensor_name(self, tensor):
        if tensor in self.tensor_name:
            name = self.tensor_name[tensor]
        # use numbering if no name can be inferred
        else:
            name = type(tensor).__name__ + str(self.name_idx)
            self.tensor_name[tensor] = name
            self.name_idx += 1
        return name

    def collect_tensor(self):
        """Collect all tensor objects tracked by python
        NOTICE:
            - the buffers for backward which is implemented in C++ are
            not tracked by python's reference counting.
            - the gradients(.grad) of Parameters is not collected, and
            I don't know why.
        """
        # FIXME: make the grad tensor collected by gc
        objects = gc.get_objects()
        tensors = [obj for obj in objects if isinstance(obj, torch.Tensor)]
        for t in tensors:
            self.device_mapping[str(t.device)].append(t)

    def get_stats(self):
        """Get the memory stat of tensors and then release them
        As a memory profiler, we cannot hold the reference to any tensors, which
        causes possibly inaccurate memory usage stats, so we delete the tensors after
        getting required stats"""
        visited_data = {}
        self.device_tensor_stat.clear()

        def get_tensor_stat(tensor):
            """Get the stat of a single tensor
            Returns:
                - stat: a tuple containing (tensor_name, tensor_size,
            tensor_numel, tensor_memory)
            """
            assert isinstance(tensor, torch.Tensor)

            name = self._get_tensor_name(tensor)

            numel = tensor.numel()
            element_size = tensor.element_size()
            fact_numel = tensor.storage().size()
            fact_memory_size = fact_numel * element_size
            # since pytorch allocate at least 512 Bytes for any tensor, round
            # up to a multiple of 512
            memory_size = math.ceil(fact_memory_size / PYTORCH_MIN_ALLOCATE) * PYTORCH_MIN_ALLOCATE

            # tensor.storage should be the actual object related to memory
            # allocation
            data_ptr = tensor.storage().data_ptr()
            if data_ptr in visited_data:
                name = '{}(->{})'.format(
                    name,
                    visited_data[data_ptr],
                )
                # don't count the memory for reusing same underlying storage
                memory_size = 0
            else:
                visited_data[data_ptr] = name

            size = tuple(tensor.size())
            # torch scalar has empty size
            if not size:
                size = (1,)

            return (name, size, numel, memory_size)

        for device, tensors in self.device_mapping.items():
            tensor_stats = []
            for tensor in tensors:

                if tensor.numel() == 0:
                    continue
                stat = get_tensor_stat(tensor)  # (name, shape, numel, memory_size)
                tensor_stats.append(stat)
                if isinstance(tensor, torch.nn.Parameter):
                    if tensor.grad is not None:
                        # manually specify the name of gradient tensor
                        self.tensor_name[tensor.grad] = '{}.grad'.format(
                            self._get_tensor_name(tensor)
                        )
                        stat = get_tensor_stat(tensor.grad)
                        tensor_stats.append(stat)

            self.device_tensor_stat[device] = tensor_stats

        self.device_mapping.clear()

    def print_stats(self, verbose=False):
        # header
        show_reuse = verbose
        template_format = '{:<40s}{:>20s}{:>10s}'
        print(template_format.format('Element type', 'Size', 'Used MEM'))
        for device, tensor_stats in self.device_tensor_stat.items():
            print('-' * LEN)
            print('Storage on {}'.format(device))
            total_mem = 0
            total_numel = 0
            for stat in tensor_stats:
                name, size, numel, mem = stat
                if not show_reuse:
                    name = name.split('(')[0]
                print(template_format.format(
                    str(name),
                    str(size),
                    readable_size(mem),
                ))
                total_mem += mem
                total_numel += numel

            print('-'*LEN)
            print('Total Tensors: {} \tUsed Memory: {}'.format(
                total_numel, readable_size(total_mem),
            ))

            if device != torch.device('cpu'):
                # with torch.cuda.device(device): NOTE not supported in
                memory_allocated = torch.cuda.memory_allocated()
                print('The allocated memory on {}: {}'.format(
                    device, readable_size(memory_allocated),
                ))
                if memory_allocated != total_mem:
                    print('Memory differs due to the matrix alignment or'
                          ' invisible gradient buffer tensors')
            print('-'*LEN)

    def collect_stats(self):
        self.collect_tensor()
        self.get_stats()
        for _, tensor_stats in self.device_tensor_stat.items():
            total_mem = 0
            for stat in tensor_stats:
                total_mem += stat[3]

            return total_mem

    def report(self, verbose=False):
        """Interface for end-users to directly print the memory usage
        args:
            - verbose: flag to show tensor.storage reuse information
        """
        self.collect_tensor()
        self.get_stats()
        self.print_stats(verbose)


@pytest.mark.parametrize('coupling', ['additive', 'affine'])
@pytest.mark.parametrize('implementation_fwd', [0, 1])
@pytest.mark.parametrize('keep_input', [True, False])
@pytest.mark.parametrize('device', ['cpu', 'cuda'])
def test_memory_saving(device, coupling, implementation_fwd, keep_input):
    """Test memory saving of the reversible block

    * tests fitting a large number of images by creating a deep network requiring large
      intermediate feature maps for training

    * keep_input = False should use less memory than keep_input = True on both GPU and CPU RAM

    * input size in bytes:            np.prod((2, 10, 10, 10)) * 4 / 1024.0 =  7.8125 kB
      for a depth=5 this yields                                  7.8125 * 5 = 39.0625 kB

    Notes
    -----
    * CPU RAM estimates uses the MemReporter from https://github.com/Stonesjtu/pytorch_memlab which tries to estimate
      torch CPU RAM usage by collecting torch.Tensors from the gc. This reliably finds the difference between
      keep_input=(True|False), however it greatly underestimates the total memory consumptions w.r.t. the GPU RAM
      estimates using PyTorch cuda module. It appears to be missing some tensors.
      FIXME CPU Ram estimates - Sil

    """

    if device == 'cuda' and not torch.cuda.is_available():
        pytest.skip('This test requires a GPU to be available')

    mem_reporter = MemReporter()

    gc.disable()
    gc.collect()
    dims = [2, 10, 10, 10]
    depth = 5
    memuse = float(np.prod(dims + [depth, 4, ])) / float(1024 ** 2)
    xx = torch.rand(*dims, device=device, dtype=torch.float32)
    ytarget = torch.rand(*dims, device=device, dtype=torch.float32)

    class SubModule(torch.nn.Module):
        def __init__(self):
            super(SubModule, self).__init__()
            self.bn = torch.nn.BatchNorm2d(10 // 2)
            self.conv = torch.nn.Conv2d(10 // 2, 10 // 2, (3, 3), padding=1)

        def forward(self, x):
            return self.bn(self.conv(x))

    class SubModuleStack(torch.nn.Module):
        def __init__(self, gm, coupling='additive', depth=10, implementation_fwd=1,
                     implementation_bwd=1, keep_input=False):
            super(SubModuleStack, self).__init__()
            self.stack = torch.nn.Sequential(
                *[revop.ReversibleBlock(gm, gm, coupling=coupling, implementation_fwd=implementation_fwd,
                                        implementation_bwd=implementation_bwd, adapter=AffineAdapterNaive,
                                        keep_input=keep_input) for _ in range(depth)]
            )

        def forward(self, x):
            return self.stack(x)

    implementation_bwd = 1
    # same convolution test
    network = SubModuleStack(SubModule(), depth=depth, keep_input=keep_input, coupling=coupling,
                             implementation_fwd=implementation_fwd, implementation_bwd=implementation_bwd)
    network.to(device)
    network.train()
    network.zero_grad()
    optim = torch.optim.RMSprop(network.parameters())
    optim.zero_grad()
    mem_start = 0 if not device == 'cuda' else \
        torch.cuda.memory_allocated() / float(1024 ** 2)

    y = network(xx)
    gc.collect()
    mem_after_forward = mem_reporter.collect_stats() / float(1024 ** 2) if not device == 'cuda' else \
        torch.cuda.memory_allocated() / float(1024 ** 2)
    loss = torch.nn.MSELoss()(y, ytarget)
    optim.zero_grad()
    loss.backward()
    optim.step()
    gc.collect()
    # mem_after_backward = mem_reporter.collect_stats() / float(1024 ** 2) if not device == 'cuda' else \
    #     torch.cuda.memory_allocated() / float(1024 ** 2)
    gc.enable()

    if device == 'cpu':
        memuse = 0.02

    if keep_input:
        assert mem_after_forward - mem_start >= memuse
    else:
        assert mem_after_forward - mem_start < (1 if device == 'cuda' else memuse)
    # assert math.floor(mem_after_backward - mem_start) >= 9



import pytest
from memcnn.train import run_experiment, main
import os
import sys
import torch


def test_main(tmp_path):
    sys.argv = ['train.py', 'cifar10', 'resnet34', '--fresh', '--no-cuda', '--workers=0']
    data_dir = str(tmp_path / "tmpdata")
    results_dir = str(tmp_path / "resdir")
    os.makedirs(data_dir)
    os.makedirs(results_dir)
    with pytest.raises(KeyError):
        main(data_dir=data_dir, results_dir=results_dir)


def dummy_dataloaders(*args, **kwargs):
    return None, None


def dummy_trainer(manager, *args, **kwargs):
    manager.save_train_state(2)


class DummyDataset(object):
    def __init__(self, *args, **kwargs):
        pass


class DummyModel(torch.nn.Module):
    def __init__(self, block):
        super(DummyModel, self).__init__()
        self.block = block
        self.conv = torch.nn.Conv2d(1, 1, 1)

    def forward(self, x):
        return self.conv(x)


content = '{' \
           '    "testsetup":' \
           '    {' \
           '        "data_loader_params": {},' \
           '        "model": "memcnn.trainers.tests.test_train.DummyModel",' \
           '        "model_params": {' \
           '            "block":"memcnn.trainers.tests.test_train.DummyDataset"' \
           '        },' \
           '        "optimizer": "torch.optim.SGD",' \
           '        "optimizer_params": {' \
           '            "lr":0.1' \
           '        },' \
           '        "trainer": "memcnn.trainers.tests.test_train.dummy_trainer",' \
           '        "trainer_params":{' \
           '            "loss":"memcnn.trainers.tests.test_train.DummyDataset"' \
           '        },' \
           '        "data_loader": "memcnn.trainers.tests.test_train.dummy_dataloaders",' \
           '        "data_loader_params": {"dataset": "memcnn.trainers.tests.test_train.DummyDataset", "workers":0}' \
           '    }' \
           '}'


def test_run_experiment(tmp_path):
    exptags = ['testsetup']
    exp_file2 = str(tmp_path / "exp_file2")
    data_dir = str(tmp_path / "tmpdata")
    results_dir = str(tmp_path / "resdir")
    with open(exp_file2, 'w') as f2:
        f2.write(content)

    start_fresh = True
    use_cuda = False
    workers = None

    with pytest.raises(RuntimeError):
        run_experiment(exptags, data_dir, results_dir, start_fresh, use_cuda, workers, exp_file2)
    os.makedirs(data_dir)
    with pytest.raises(RuntimeError):
        run_experiment(exptags, data_dir, results_dir, start_fresh, use_cuda, workers, exp_file2)
    os.makedirs(results_dir)
    run_experiment(exptags, data_dir, results_dir, start_fresh, use_cuda, workers, exp_file2)
    run_experiment(exptags, data_dir, results_dir, False, use_cuda, workers, exp_file2)



# -*- coding: utf-8 -*-

import torch.nn as nn
import warnings
from memcnn.models.additive import AdditiveBlock
from memcnn.models.affine import AffineBlock
from memcnn.models.utils import pytorch_version_one_and_above


warnings.filterwarnings(action='ignore', category=UserWarning)


class ReversibleBlock(nn.Module):
    def __init__(self, Fm, Gm=None, coupling='additive', keep_input=False, keep_input_inverse=False,
                 implementation_fwd=1, implementation_bwd=1, adapter=None):
        """The ReversibleBlock

        Note
        ----
        The `implementation_fwd` and `implementation_bwd` parameters can be set to one of the following implementations:

        * -1 Naive implementation without reconstruction on the backward pass.
        * 0  Memory efficient implementation, compute gradients directly.
        * 1  Memory efficient implementation, similar to approach in Gomez et al. 2017.


        Parameters
        ----------
            Fm : :obj:`torch.nn.Module`
                A torch.nn.Module encapsulating an arbitrary function

            Gm : :obj:`torch.nn.Module`, optional
                A torch.nn.Module encapsulating an arbitrary function
                (If not specified a deepcopy of Fm is used as a Module)

            coupling : :obj:`str`, optional
                Type of coupling ['additive', 'affine']. Default = 'additive'

            keep_input : :obj:`bool`, optional
                Set to retain the input information on forward, by default it can be discarded since it will be
                reconstructed upon the backward pass.

            keep_input_inverse : :obj:`bool`, optional
                Set to retain the input information on inverse, by default it can be discarded since it will be
                reconstructed upon the backward pass.

            implementation_fwd : :obj:`int`, optional
                Switch between different Operation implementations for forward training (Default = 1).
                If using the naive implementation (-1) then `keep_input` should be True.

            implementation_bwd : :obj:`int`, optional
                Switch between different Operation implementations for backward training (Default = 1).
                If using the naive implementation (-1) then `keep_input_inverse` should be True.

            adapter : :obj:`class`, optional
                Only relevant when using the 'affine' coupling.
                Should be a class of type :obj:`torch.nn.Module` that serves as an
                optional wrapper class A for Fm and Gm which must output
                s, t = A(x) with shape(s) = shape(t) = shape(x).
                s, t are respectively the scale and shift tensors for the affine coupling.

        Attributes
        ----------
            keep_input : :obj:`bool`, optional
                Set to retain the input information on forward, by default it can be discarded since it will be
                reconstructed upon the backward pass.

            keep_input_inverse : :obj:`bool`, optional
                Set to retain the input information on inverse, by default it can be discarded since it will be
                reconstructed upon the backward pass.

        Raises
        ------
        NotImplementedError
            If an unknown coupling or implementation is given.

        """
        super(ReversibleBlock, self).__init__()
        self.keep_input = keep_input
        self.keep_input_inverse = keep_input_inverse
        if coupling == 'additive':
            self.rev_block = AdditiveBlock(Fm, Gm,
                                           implementation_fwd=implementation_fwd, implementation_bwd=implementation_bwd)
        elif coupling == 'affine':
            self.rev_block = AffineBlock(Fm, Gm, adapter=adapter,
                                         implementation_fwd=implementation_fwd, implementation_bwd=implementation_bwd)
        else:
            raise NotImplementedError('Unknown coupling method: %s' % coupling)

    def forward(self, x):
        """Forward operation :math:`R(x) = y`

        Parameters
        ----------
            x : :obj:`torch.Tensor`
                Input torch tensor.

        Returns
        -------
            :obj:`torch.Tensor`
                Output torch tensor y.

        """
        y = self.rev_block(x)
        # clears the referenced storage data linked to the input tensor as it can be reversed on the backward pass
        if not self.keep_input:
            if not pytorch_version_one_and_above:
                # PyTorch 0.4 way to clear storage
                x.data.set_()
            else:
                # PyTorch 1.0+ way to clear storage
                x.storage().resize_(0)

        return y

    def inverse(self, y):
        """Inverse operation :math:`R^{-1}(y) = x`

        Parameters
        ----------
            y : :obj:`torch.Tensor`
                Input torch tensor.

        Returns
        -------
            :obj:`torch.Tensor`
                Output torch tensor x.

        """
        x = self.rev_block.inverse(y)
        # clears the referenced storage data linked to the input tensor as it can be reversed on the backward pass
        if not self.keep_input_inverse:
            if not pytorch_version_one_and_above:
                # PyTorch 0.4 way to clear storage
                y.data.set_()
            else:
                # PyTorch 1.0+ way to clear storage
                y.storage().resize_(0)

        return x



import collections
import tensorflow as tf
import glob
import json
import os


def parse_logs(dir_path, output_file = None):
    # collect all log_files from dir_path
    log_files = glob.glob(os.path.join(dir_path, 'events.out.tfevents.*'))
    event_dict = {}
    # iterate all log_fiiles in order of creation
    for log_file in log_files:
        for event in tf.train.summary_iterator(log_file):
            for value in event.summary.value:
                if not (value.tag in event_dict):
                    event_dict[value.tag] = collections.OrderedDict()
                if value.HasField('simple_value'):
                    event_dict[value.tag][event.step] = (event.wall_time, event.step, value.simple_value)
    # flatten ordered dicts to lists
    for key in event_dict.keys():
        event_dict[key] = list(event_dict[key].values())
    # optionally write event_dict to json
    if output_file is not None:
        with open(output_file, 'w') as f:
            json.dump(event_dict, f)
    return event_dict



import pytest
import torch
import torch.nn
import numpy as np
import copy
from memcnn.models.affine import AffineAdapterNaive, AffineAdapterSigmoid
from memcnn import ReversibleBlock


def set_seeds(seed):
    np.random.seed(seed)
    torch.manual_seed(seed)


@pytest.mark.parametrize('coupling', ['additive', 'affine'])
def test_reversible_block_additive_notimplemented(coupling):
    fm = torch.nn.Conv2d(10, 10, (3, 3), padding=1)
    X = torch.zeros(1, 20, 10, 10)
    with pytest.raises(NotImplementedError):
        f = ReversibleBlock(fm, coupling=coupling, implementation_bwd=0, implementation_fwd=-2,
                                  adapter=AffineAdapterNaive)
        f.forward(X)
    with pytest.raises(NotImplementedError):
        f = ReversibleBlock(fm, coupling=coupling, implementation_bwd=-2, implementation_fwd=0,
                                  adapter=AffineAdapterNaive)
        f.inverse(X)
    with pytest.raises(NotImplementedError):
        ReversibleBlock(fm, coupling='unknown', implementation_bwd=-2, implementation_fwd=0,
                              adapter=AffineAdapterNaive)


@pytest.mark.parametrize('coupling,adapter', [('additive', None),
                                              ('affine', AffineAdapterNaive),
                                              ('affine', AffineAdapterSigmoid)])
def test_reversible_block_fwd_bwd(coupling, adapter):
    """ReversibleBlock test of the memory saving forward and backward passes

    * test inversion Y = RB(X) and X = RB.inverse(Y)
    * test training the block for a single step and compare weights for implementations: 0, 1
    * test automatic discard of input X and its retrieval after the backward pass
    * test usage of BN to identify non-contiguous memory blocks

    """
    dims = (2, 10, 8, 8)
    data = np.random.random(dims).astype(np.float32)
    target_data = np.random.random(dims).astype(np.float32)

    class SubModule(torch.nn.Module):
        def __init__(self, in_filters, out_filters):
            super(SubModule, self).__init__()
            self.bn = torch.nn.BatchNorm2d(out_filters)
            self.conv = torch.nn.Conv2d(in_filters, out_filters, (3, 3), padding=1)

        def forward(self, x):
            return self.bn(self.conv(x))

    Gm = SubModule(in_filters=5, out_filters=5 if coupling == 'additive' or adapter is AffineAdapterNaive else 10)

    s_grad = [p.data.numpy().copy() for p in Gm.parameters()]
    for seed in range(10):
        set_seeds(seed)
        for bwd in [False, True]:
            impl_out, impl_grad = [], []
            for keep_input_sub in [False, True]:
                for keep_input_inverse_sub in [False, True]:
                    for implementation_fwd in [-1, 0, 1]:
                        for implementation_bwd in [-1, 0, 1]:
                            keep_input = keep_input_sub or implementation_fwd == -1
                            keep_input_inverse = keep_input_inverse_sub or implementation_bwd == -1
                            # print(bwd, coupling, keep_input, implementation_fwd, implementation_bwd)
                            # test with zero padded convolution
                            X = torch.from_numpy(data.copy())
                            Ytarget = torch.from_numpy(target_data.copy())
                            Xshape = X.shape
                            Gm2 = copy.deepcopy(Gm)
                            rb = ReversibleBlock(Gm2, coupling=coupling, implementation_fwd=implementation_fwd,
                                                       implementation_bwd=implementation_bwd, adapter=adapter,
                                                       keep_input=keep_input, keep_input_inverse=keep_input_inverse)
                            rb.train()
                            rb.zero_grad()

                            optim = torch.optim.RMSprop(rb.parameters())
                            optim.zero_grad()
                            if not bwd:
                                Xin = X.clone()
                                Y = rb(Xin)
                                Yrev = Y.clone()
                                Xinv = rb.inverse(Yrev)
                            else:
                                Xin = X.clone()
                                Y = rb.inverse(Xin)
                                Yrev = Y.clone()
                                Xinv = rb(Yrev)
                            loss = torch.nn.MSELoss()(Y, Ytarget)

                            # has input been retained/discarded after forward (and backward) passes?

                            def test_memory_cleared(var, isclear, shape):
                                if isclear:
                                    assert var.storage().size() == 0
                                else:
                                    assert var.storage().size() > 0
                                    assert var.shape == shape

                            if not bwd:
                                test_memory_cleared(Xin, not keep_input, Xshape)
                                test_memory_cleared(Yrev, not keep_input_inverse, Xshape)
                            else:
                                test_memory_cleared(Yrev, not keep_input, Xshape)
                                test_memory_cleared(Xin, not keep_input_inverse, Xshape)

                            optim.zero_grad()
                            loss.backward()
                            optim.step()

                            assert Y.shape == Xshape
                            assert X.data.numpy().shape == data.shape
                            assert np.allclose(X.data.numpy(), data, atol=1e-06)
                            assert np.allclose(X.data.numpy(), Xinv.data.numpy(), atol=1e-05)
                            impl_out.append(Y.data.numpy().copy())
                            impl_grad.append([p.data.numpy().copy() for p in Gm2.parameters()])
                            assert not np.allclose(impl_grad[-1][0], s_grad[0])

                    # output and gradients similar over all implementations?
                    for i in range(0, len(impl_grad) - 1, 1):
                        assert np.allclose(impl_grad[i][0], impl_grad[i + 1][0])
                        assert np.allclose(impl_out[i], impl_out[i + 1])


@pytest.mark.parametrize('coupling,adapter', [('additive', None),
                                              ('affine', AffineAdapterNaive),
                                              ('affine', AffineAdapterSigmoid)])
def test_revblock_chained(coupling, adapter):
    set_seeds(42)
    dims = (2, 10, 8, 8)
    data = np.random.random(dims).astype(np.float32)
    target_data = np.random.random(dims).astype(np.float32)

    X = torch.from_numpy(data.copy())
    Ytarget = torch.from_numpy(target_data.copy())

    class SubModule(torch.nn.Module):
        def __init__(self, in_filters, out_filters):
            super(SubModule, self).__init__()
            self.bn = torch.nn.BatchNorm2d(out_filters)
            self.conv = torch.nn.Conv2d(in_filters, out_filters, (3, 3), padding=1)

        def forward(self, x):
            return self.bn(self.conv(x))

    class SubModuleStack(torch.nn.Module):
        def __init__(self, Gm, coupling='additive', depth=10, implementation_fwd=1, implementation_bwd=1,
                     keep_input=False, adapter=None):
            super(SubModuleStack, self).__init__()
            self.stack = torch.nn.Sequential(
                *[ReversibleBlock(Gm, Gm, coupling=coupling, implementation_fwd=implementation_fwd,
                                        implementation_bwd=implementation_bwd, adapter=adapter,
                                        keep_input=keep_input) for _ in range(depth)]
            )

        def forward(self, x):
            return self.stack(x)


    Gm = SubModule(in_filters=5, out_filters=5 if coupling == 'additive' or adapter is AffineAdapterNaive else 10)
    rb = SubModuleStack(Gm, coupling=coupling, depth=2, keep_input=False, adapter=adapter)
    rb.train()
    rb.zero_grad()

    optim = torch.optim.RMSprop(rb.parameters())
    optim.zero_grad()

    Xin = X.clone()
    Y = rb(Xin)

    loss = torch.nn.MSELoss()(Y, Ytarget)

    optim.zero_grad()
    loss.backward()
    optim.step()


@pytest.mark.parametrize('coupling', ['additive', 'affine'])
def test_revblock_simple_inverse(coupling):
    """ReversibleBlock inverse test

    * test inversion Y = RB(X) and X = RB.inverse(Y)

    """
    for seed in range(10):
        set_seeds(seed)
        for implementation_fwd in [-1, 0, 1]:
            for implementation_bwd in [-1, 0, 1]:
                # define some data
                X = torch.rand(2, 4, 5, 5)

                # define an arbitrary reversible function
                fn = ReversibleBlock(torch.nn.Conv2d(2, 2, 3, padding=1), keep_input=False, coupling=coupling,
                                           implementation_fwd=implementation_fwd, implementation_bwd=implementation_bwd,
                                           adapter=AffineAdapterNaive)

                # compute output
                Y = fn.forward(X.clone())

                # compute input from output
                X2 = fn.inverse(Y)

                # check that the inverted output and the original input are approximately similar
                assert np.allclose(X2.data.numpy(), X.data.numpy(), atol=1e-06)


@pytest.mark.parametrize('coupling', ['additive', 'affine'])
@pytest.mark.parametrize('implementation_fwd', [-1, 0, 1])
@pytest.mark.parametrize('implementation_bwd', [-1, 0, 1])
def test_normal_vs_revblock(coupling, implementation_fwd, implementation_bwd):
    """ReversibleBlock test if similar gradients and weights results are obtained after similar training

    * test training the block for a single step and compare weights and grads for implementations: 0, 1
    * test against normal non Reversible Block function
    * test if recreated input and produced output are contiguous

    """
    for seed in range(10):
        set_seeds(seed)

        X = torch.rand(2, 4, 5, 5)

        # define models and their copies
        c1 = torch.nn.Conv2d(2, 2, 3, padding=1)
        c2 = torch.nn.Conv2d(2, 2, 3, padding=1)
        c1_2 = copy.deepcopy(c1)
        c2_2 = copy.deepcopy(c2)

        # are weights between models the same, but do they differ between convolutions?
        assert torch.equal(c1.weight, c1_2.weight)
        assert torch.equal(c2.weight, c2_2.weight)
        assert torch.equal(c1.bias, c1_2.bias)
        assert torch.equal(c2.bias, c2_2.bias)
        assert not torch.equal(c1.weight, c2.weight)

        # define optimizers
        optim1 = torch.optim.SGD([e for e in c1.parameters()] + [e for e in c2.parameters()], 0.1)
        optim2 = torch.optim.SGD([e for e in c1_2.parameters()] + [e for e in c2_2.parameters()], 0.1)
        for e in [c1, c2, c1_2, c2_2]:
            e.train()

        # define an arbitrary reversible function and define graph for model 1
        Xin = X.clone()
        fn = ReversibleBlock(c1_2, c2_2, keep_input=False, coupling=coupling, adapter=AffineAdapterNaive,
                                   implementation_fwd=implementation_fwd, implementation_bwd=implementation_bwd)
        Y = fn.forward(Xin)
        loss2 = torch.mean(Y)

        # define the reversible function without custom backprop and define graph for model 2
        XX = X.clone().data
        XX.requires_grad = True
        x1, x2 = torch.chunk(XX, 2, dim=1)
        if coupling == 'additive':
            y1 = x1 + c1.forward(x2)
            y2 = x2 + c2.forward(y1)
        elif coupling == 'affine':
            fmr2 = c1.forward(x2)
            fmr1 = torch.exp(fmr2)
            y1 = (x1 * fmr1) + fmr2
            gmr2 = c2.forward(y1)
            gmr1 = torch.exp(gmr2)
            y2 = (x2 * gmr1) + gmr2
        else:
            raise NotImplementedError()
        YY = torch.cat([y1, y2], dim=1)

        loss = torch.mean(YY)

        # compute gradients manually
        grads = torch.autograd.grad(loss, (XX, c1.weight, c2.weight, c1.bias, c2.bias), None, retain_graph=True)

        # compute gradients and perform optimization model 2
        loss.backward()
        optim1.step()

        # gradients computed manually match those of the .backward() pass
        assert torch.equal(c1.weight.grad, grads[1])
        assert torch.equal(c2.weight.grad, grads[2])
        assert torch.equal(c1.bias.grad, grads[3])
        assert torch.equal(c2.bias.grad, grads[4])

        # weights differ after training a single model?
        assert not torch.equal(c1.weight, c1_2.weight)
        assert not torch.equal(c2.weight, c2_2.weight)
        assert not torch.equal(c1.bias, c1_2.bias)
        assert not torch.equal(c2.bias, c2_2.bias)

        # compute gradients and perform optimization model 1
        loss2.backward()
        optim2.step()

        # input is contiguous tests
        assert Xin.is_contiguous()
        assert Y.is_contiguous()

        # weights are approximately the same after training both models?
        assert np.allclose(c1.weight.data.numpy(), c1_2.weight.data.numpy(), atol=1e-06)
        assert np.allclose(c2.weight.data.numpy(), c2_2.weight.data.numpy())
        assert np.allclose(c1.bias.data.numpy(), c1_2.bias.data.numpy())
        assert np.allclose(c2.bias.data.numpy(), c2_2.bias.data.numpy())

        # gradients are approximately the same after training both models?
        assert np.allclose(c1.weight.grad.data.numpy(), c1_2.weight.grad.data.numpy(), atol=1e-06)
        assert np.allclose(c2.weight.grad.data.numpy(), c2_2.weight.grad.data.numpy())
        assert np.allclose(c1.bias.grad.data.numpy(), c1_2.bias.grad.data.numpy())
        assert np.allclose(c2.bias.grad.data.numpy(), c2_2.bias.grad.data.numpy())



# -*- coding: utf-8 -*-

"""Top-level package for MemCNN."""

__author__ = """Sil van de Leemput"""
__email__ = 'silvandeleemput@gmail.com'
__version__ = '0.3.5'


from memcnn.models.revop import ReversibleBlock

__all__ = [
    'ReversibleBlock'
]
